@@ -45,13 +45,8 @@
 #define MIN_SHADE_RESISTANCE_MIN       0
 #define MIN_SHADE_RESISTANCE_MAX       100
 
-static char *winType[] = {
-    N_("Toolbar"),
-    N_("Utility"),
-    N_("Dialog"),
-    N_("Normal")
-};
-#define N_WIN_TYPE (sizeof (winType) / sizeof (winType[0]))
+#define MIN_WINDOW_MATCH_DEFAULT \
+    "Toolbar | Utility | Dialog | Normal"
 
 static int displayPrivateIndex;
 
@@ -64,7 +59,7 @@ typedef struct _MinDisplay {
 
 #define MIN_SCREEN_OPTION_SPEED		   0
 #define MIN_SCREEN_OPTION_TIMESTEP	   1
-#define MIN_SCREEN_OPTION_WINDOW_TYPE	   2
+#define MIN_SCREEN_OPTION_WINDOW_MATCH	   2
 #define MIN_SCREEN_OPTION_SHADE_RESISTANCE 3
 #define MIN_SCREEN_OPTION_NUM		   4
 
@@ -85,8 +80,6 @@ typedef struct _MinScreen {
 
     int shadeStep;
 
-    unsigned int wMask;
-
     int moreAdjust;
 } MinScreen;
 
@@ -168,12 +161,9 @@ minSetScreenOption (CompScreen      *screen,
 	    return TRUE;
 	}
 	break;
-    case MIN_SCREEN_OPTION_WINDOW_TYPE:
-	if (compSetOptionList (o, value))
-	{
-	    ms->wMask = compWindowTypeMaskFromStringList (&o->value);
+    case MIN_SCREEN_OPTION_WINDOW_MATCH:
+	if (compSetMatchOption (o, value))
 	    return TRUE;
-	}
 	break;
     case MIN_SCREEN_OPTION_SHADE_RESISTANCE:
 	if (compSetIntOption (o, value))
@@ -196,7 +186,6 @@ static void
 minScreenInitOptions (MinScreen *ms)
 {
     CompOption *o;
-    int	       i;
 
     o = &ms->opt[MIN_SCREEN_OPTION_SPEED];
     o->name		= "speed";
@@ -218,21 +207,14 @@ minScreenInitOptions (MinScreen *ms)
     o->rest.f.max	= MIN_TIMESTEP_MAX;
     o->rest.f.precision = MIN_TIMESTEP_PRECISION;
 
-    o = &ms->opt[MIN_SCREEN_OPTION_WINDOW_TYPE];
-    o->name	         = "window_types";
-    o->shortDesc         = N_("Window Types");
-    o->longDesc	         = N_("Window types that should be transformed when "
-	"minimized");
-    o->type	         = CompOptionTypeList;
-    o->value.list.type   = CompOptionTypeString;
-    o->value.list.nValue = N_WIN_TYPE;
-    o->value.list.value  = malloc (sizeof (CompOptionValue) * N_WIN_TYPE);
-    for (i = 0; i < N_WIN_TYPE; i++)
-	o->value.list.value[i].s = strdup (winType[i]);
-    o->rest.s.string     = windowTypeString;
-    o->rest.s.nString    = nWindowTypeString;
+    o = &ms->opt[MIN_SCREEN_OPTION_WINDOW_MATCH];
+    o->name	 = "window_match";
+    o->shortDesc = N_("Minimize Windows");
+    o->longDesc	 = N_("Windows that should be transformed when minimized");
+    o->type	 = CompOptionTypeMatch;
 
-    ms->wMask = compWindowTypeMaskFromStringList (&o->value);
+    matchInit (&o->value.match);
+    matchAddFromString (&o->value.match, MIN_WINDOW_MATCH_DEFAULT);
 
     o = &ms->opt[MIN_SCREEN_OPTION_SHADE_RESISTANCE];
     o->name		= "shade_resistance";
@@ -272,7 +254,7 @@ minSetShade (CompWindow *w,
     w->matrix.x0 -= (w->attrib.x * w->matrix.xx);
     w->matrix.y0 -= ((w->attrib.y - (h - shade)) * w->matrix.yy);
 
-    (*w->screen->windowResizeNotify) (w);
+    (*w->screen->windowResizeNotify) (w, 0, 0, 0, 0);
 }
 
 static Bool
@@ -361,7 +343,7 @@ adjustMinVelocity (CompWindow *w)
 	xScale = yScale = 1.0f;
     }
 
-    dx = x1 - w->attrib.x;
+    dx = x1 - (w->attrib.x + mw->tx);
 
     adjust = dx * 0.15f;
     amount = fabs (dx) * 1.5f;
@@ -372,7 +354,7 @@ adjustMinVelocity (CompWindow *w)
 
     mw->xVelocity = (amount * mw->xVelocity + adjust) / (amount + 1.0f);
 
-    dy = y1 - w->attrib.y;
+    dy = y1 - (w->attrib.y + mw->ty);
 
     adjust = dy * 0.15f;
     amount = fabs (dy) * 1.5f;
@@ -434,7 +416,7 @@ minPreparePaintScreen (CompScreen *s,
     if (ms->moreAdjust)
     {
 	CompWindow *w;
-	int        steps, dx, dy, h;
+	int        steps, h;
 	float      amount, chunk;
 
 	amount = msSinceLastPaint * 0.05f * ms->speed;
@@ -461,13 +443,6 @@ minPreparePaintScreen (CompScreen *s,
 		    mw->xScale += mw->xScaleVelocity * chunk;
 		    mw->yScale += mw->yScaleVelocity * chunk;
 
-		    dx = (w->serverX + mw->tx) - w->attrib.x;
-		    dy = (w->serverY + mw->ty) - w->attrib.y;
-
-		    moveWindow (w, dx, dy, FALSE, FALSE);
-
-		    (*s->setWindowScale) (w, mw->xScale, mw->yScale);
-
 		    if (!mw->adjust)
 		    {
 			mw->state = mw->newState;
@@ -597,6 +572,7 @@ minDonePaintScreen (CompScreen *s)
 static Bool
 minPaintScreen (CompScreen		*s,
 		const ScreenPaintAttrib *sAttrib,
+		const CompTransform	*transform,
 		Region		        region,
 		int			output,
 		unsigned int		mask)
@@ -609,7 +585,7 @@ minPaintScreen (CompScreen		*s,
 	mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;
 
     UNWRAP (ms, s, paintScreen);
-    status = (*s->paintScreen) (s, sAttrib, region, output, mask);
+    status = (*s->paintScreen) (s, sAttrib, transform, region, output, mask);
     WRAP (ms, s, paintScreen, minPaintScreen);
 
     return status;
@@ -618,6 +594,7 @@ minPaintScreen (CompScreen		*s,
 static Bool
 minPaintWindow (CompWindow		*w,
 		const WindowPaintAttrib *attrib,
+		const CompTransform	*transform,
 		Region			region,
 		unsigned int		mask)
 {
@@ -628,11 +605,48 @@ minPaintWindow (CompWindow		*w,
     MIN_WINDOW (w);
 
     if (mw->adjust)
-	mask |= PAINT_WINDOW_TRANSFORMED_MASK;
+    {
+	FragmentAttrib fragment;
+	CompTransform  wTransform = *transform;
 
-    UNWRAP (ms, s, paintWindow);
-    status = (*s->paintWindow) (w, attrib, region, mask);
-    WRAP (ms, s, paintWindow, minPaintWindow);
+	if (mask & PAINT_WINDOW_OCCLUSION_DETECTION_MASK)
+	    return FALSE;
+
+	UNWRAP (ms, s, paintWindow);
+	status = (*s->paintWindow) (w, attrib, transform, region,
+				    mask | PAINT_WINDOW_NO_CORE_INSTANCE_MASK);
+	WRAP (ms, s, paintWindow, minPaintWindow);
+
+	initFragmentAttrib (&fragment, &w->lastPaint);
+
+	if (w->alpha || fragment.opacity != OPAQUE)
+	    mask |= PAINT_WINDOW_TRANSLUCENT_MASK;
+
+	matrixTranslate (&wTransform, w->attrib.x, w->attrib.y, 0.0f);
+	matrixScale (&wTransform, mw->xScale, mw->yScale, 0.0f);
+	matrixTranslate (&wTransform,
+			 mw->tx / mw->xScale - w->attrib.x,
+			 mw->ty / mw->yScale - w->attrib.y,
+			 0.0f);
+
+	glPushMatrix ();
+	glLoadMatrixf (wTransform.m);
+
+	(*s->drawWindow) (w, &wTransform, &fragment, region,
+			  mask | PAINT_WINDOW_TRANSFORMED_MASK);
+
+	glPopMatrix ();
+    }
+    else
+    {
+	/* no core instance from windows that have been animated */
+	if (mw->state == IconicState)
+	    mask |= PAINT_WINDOW_NO_CORE_INSTANCE_MASK;
+
+	UNWRAP (ms, s, paintWindow);
+	status = (*s->paintWindow) (w, attrib, transform, region, mask);
+	WRAP (ms, s, paintWindow, minPaintWindow);
+    }
 
     return status;
 }
@@ -671,8 +685,11 @@ minHandleEvent (CompDisplay *d,
 	{
 	    MIN_SCREEN (w->screen);
 
-	    if (w->pendingUnmaps) /* Normal -> Iconic */
+	    if (w->pendingUnmaps && onCurrentDesktop (w)) /* Normal -> Iconic */
 	    {
+		CompMatch *match =
+		    &ms->opt[MIN_SCREEN_OPTION_WINDOW_MATCH].value.match;
+
 		MIN_WINDOW (w);
 
 		if (w->shaded)
@@ -699,12 +716,12 @@ minHandleEvent (CompDisplay *d,
 			addWindowDamage (w);
 		    }
 		}
-		else if (!w->invisible && (ms->wMask & w->type))
+		else if (!w->invisible && matchEval (match, w))
 		{
-		    mw->newState = IconicState;
-
 		    if (minGetWindowIconGeometry (w, &mw->icon))
 		    {
+			mw->newState = IconicState;
+
 			mw->xScale = w->paint.xScale;
 			mw->yScale = w->paint.yScale;
 			mw->tx	   = w->attrib.x - w->serverX;
@@ -726,10 +743,6 @@ minHandleEvent (CompDisplay *d,
 
 			addWindowDamage (w);
 		    }
-		    else
-		    {
-			mw->state = mw->newState;
-		    }
 		}
 	    }
 	    else  /* X -> Withdrawn */
@@ -750,13 +763,6 @@ minHandleEvent (CompDisplay *d,
 			XDestroyRegion (mw->region);
 			mw->region = NULL;
 		    }
-
-		    (*w->screen->setWindowScale) (w, 1.0f, 1.0f);
-
-		    moveWindow (w,
-				w->serverX - w->attrib.x,
-				w->serverY - w->attrib.y,
-				FALSE, TRUE);
 		}
 
 		mw->state = NormalState;
@@ -776,18 +782,22 @@ minDamageWindowRect (CompWindow *w,
 		     Bool	initial,
 		     BoxPtr     rect)
 {
-    Bool status;
+    Bool status = FALSE;
 
     MIN_SCREEN (w->screen);
+    MIN_WINDOW (w);
 
     if (initial)
     {
-	MIN_WINDOW (w);
-
 	if (mw->state == IconicState)
 	{
-	    if (!w->invisible	      &&
-		(ms->wMask & w->type) &&
+	    CompMatch *match =
+		&ms->opt[MIN_SCREEN_OPTION_WINDOW_MATCH].value.match;
+
+	    mw->state = NormalState;
+
+	    if (!w->invisible	     &&
+		matchEval (match, w) &&
 		minGetWindowIconGeometry (w, &mw->icon))
 	    {
 		if (!mw->adjust)
@@ -800,25 +810,9 @@ minDamageWindowRect (CompWindow *w,
 		    mw->xScale = (float) mw->icon.width  / w->width;
 		    mw->yScale = (float) mw->icon.height / w->height;
 
-		    moveWindow (w,
-				mw->icon.x - w->attrib.x,
-				mw->icon.y - w->attrib.y,
-				FALSE, TRUE);
-
-		    (*w->screen->setWindowScale) (w, mw->xScale, mw->yScale);
-
 		    addWindowDamage (w);
 		}
 	    }
-	    else
-	    {
-		moveWindow (w,
-			    w->serverX - w->attrib.x,
-			    w->serverY - w->attrib.y,
-			    FALSE, TRUE);
-
-		(*w->screen->setWindowScale) (w, 1.0f, 1.0f);
-	    }
 	}
 	else if (mw->region && mw->shade < w->height)
 	{
@@ -828,7 +822,7 @@ minDamageWindowRect (CompWindow *w,
 		XOffsetRegion (mw->region, -w->attrib.x, -w->attrib.y);
 
 		/* bind pixmap here so we have something to unshade with */
-		if (!w->texture->pixmap)
+		if (!w->texture->pixmap && !w->bindFailed)
 		    bindWindow (w);
 
 		ms->moreAdjust = TRUE;
@@ -841,9 +835,20 @@ minDamageWindowRect (CompWindow *w,
 
 	mw->newState = NormalState;
     }
+    else if (mw->adjust)
+    {
+	damageTransformedWindowRect (w,
+				     mw->xScale,
+				     mw->yScale,
+				     mw->tx,
+				     mw->ty,
+				     rect);
+
+	status = TRUE;
+    }
 
     UNWRAP (ms, w->screen, damageWindowRect);
-    status = (*w->screen->damageWindowRect) (w, initial, rect);
+    status |= (*w->screen->damageWindowRect) (w, initial, rect);
     WRAP (ms, w->screen, damageWindowRect, minDamageWindowRect);
 
     return status;
@@ -935,6 +940,9 @@ minInitScreen (CompPlugin *p,
 
     minScreenInitOptions (ms);
 
+    matchUpdate (s->display,
+		 &ms->opt[MIN_SCREEN_OPTION_WINDOW_MATCH].value.match);
+
     WRAP (ms, s, preparePaintScreen, minPreparePaintScreen);
     WRAP (ms, s, donePaintScreen, minDonePaintScreen);
     WRAP (ms, s, paintScreen, minPaintScreen);
@@ -953,6 +961,8 @@ minFiniScreen (CompPlugin *p,
 {
     MIN_SCREEN (s);
 
+    matchFini (&ms->opt[MIN_SCREEN_OPTION_WINDOW_MATCH].value.match);
+
     freeWindowPrivateIndex (s, ms->windowPrivateIndex);
 
     UNWRAP (ms, s, preparePaintScreen);
@@ -985,17 +995,26 @@ minInitWindow (CompPlugin *p,
 
     mw->unmapCnt = 0;
 
-    if (w->shaded)
+    if (w->state & CompWindowStateHiddenMask)
     {
-	mw->state = mw->newState = NormalState;
-	mw->shade = 0;
-	mw->region = XCreateRegion ();
+	if (w->shaded)
+	{
+	    mw->state = mw->newState = NormalState;
+	    mw->shade = 0;
+	    mw->region = XCreateRegion ();
+	}
+	else
+	{
+	    mw->state = mw->newState = minGetWindowState (w);
+	    mw->shade = MAXSHORT;
+	    mw->region = NULL;
+	}
     }
     else
     {
-	mw->state = mw->newState = minGetWindowState (w);
-	mw->shade = MAXSHORT;
-	mw->region = NULL;
+	    mw->state = mw->newState = NormalState;
+	    mw->shade = MAXSHORT;
+	    mw->region = NULL;
     }
 
     w->privates[ms->windowPrivateIndex].ptr = mw;
