@@ -76,8 +76,6 @@ typedef struct _PlaneDisplay {
     HandleEventProc	handleEvent;
 
     CompOption		opt[PLANE_N_DISPLAY_OPTIONS];
-
-    int			flipTime;
 } PlaneDisplay;
 
 typedef struct _PlaneScreen {
@@ -86,7 +84,6 @@ typedef struct _PlaneScreen {
     DonePaintScreenProc			donePaintScreen;
     PaintScreenProc			paintScreen;
 
-    SetScreenOptionForPluginProc	setScreenOptionForPlugin;
     WindowGrabNotifyProc		windowGrabNotify;
     WindowUngrabNotifyProc		windowUngrabNotify;
 
@@ -465,22 +462,6 @@ planeWindowUngrabNotify (CompWindow *w)
     WRAP (ps, w->screen, windowUngrabNotify, planeWindowUngrabNotify);
 }
 
-static Bool
-planeSetScreenOptionForPlugin (CompScreen      *s,
-			       char	       *plugin,
-			       char	       *name,
-			       CompOptionValue *value)
-{
-    PLANE_SCREEN (s);
-    Bool status;
-
-    UNWRAP (ps, s, setScreenOptionForPlugin);
-    status = (*s->setScreenOptionForPlugin) (s, plugin, name, value);
-    WRAP (ps, s, setScreenOptionForPlugin, planeSetScreenOptionForPlugin);
-
-    return status;
-}
-
 static CompOption *
 planeGetDisplayOptions (CompDisplay *display,
 			int	    *count)
@@ -825,7 +806,6 @@ planeInitScreen (CompPlugin *p,
     WRAP (ps, s, preparePaintScreen, planePreparePaintScreen);
     WRAP (ps, s, donePaintScreen, planeDonePaintScreen);
     WRAP (ps, s, paintScreen, planePaintScreen);
-    WRAP (ps, s, setScreenOptionForPlugin, planeSetScreenOptionForPlugin);
     WRAP (ps, s, windowGrabNotify, planeWindowGrabNotify);
     WRAP (ps, s, windowUngrabNotify, planeWindowUngrabNotify);
 
@@ -839,12 +819,30 @@ planeFiniScreen (CompPlugin *p,
 		 CompScreen *s)
 {
     PLANE_SCREEN (s);
+    PLANE_DISPLAY (s->display);
+
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_LEFT].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_RIGHT].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_DOWN].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_UP].value.action);
+
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_1].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_2].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_3].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_4].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_5].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_6].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_7].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_8].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_9].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_10].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_11].value.action);
+    removeScreenAction (s, &pd->opt[PLANE_DISPLAY_OPTION_TO_12].value.action);
 
     UNWRAP (ps, s, paintTransformedScreen);
     UNWRAP (ps, s, preparePaintScreen);
     UNWRAP (ps, s, donePaintScreen);
     UNWRAP (ps, s, paintScreen);
-    UNWRAP (ps, s, setScreenOptionForPlugin);
     UNWRAP (ps, s, windowGrabNotify);
     UNWRAP (ps, s, windowUngrabNotify);
 
