@@ -509,7 +509,18 @@ rotatePreparePaintScreen (CompScreen *s,
 			}
 		    }
 		    else
-			focusDefaultWindow (s->display);
+		    {
+			int i;
+
+			for (i = 0; i < s->maxGrab; i++)
+			    if (s->grabs[i].active &&
+				strcmp ("switcher", s->grabs[i].name) == 0)
+				break;
+
+			/* only focus default window if switcher isn't active */
+			if (i == s->maxGrab)
+			    focusDefaultWindow (s->display);
+		    }
 
 		    rs->moveWindow = 0;
 		}
@@ -2100,6 +2111,20 @@ rotateFiniScreen (CompPlugin *p,
 		  CompScreen *s)
 {
     ROTATE_SCREEN (s);
+    ROTATE_DISPLAY (s->display);
+
+    removeScreenAction (s, 
+			&rd->opt[ROTATE_DISPLAY_OPTION_INITIATE].value.action);
+    removeScreenAction (s, &rd->opt[ROTATE_DISPLAY_OPTION_LEFT].value.action);
+    removeScreenAction (s, &rd->opt[ROTATE_DISPLAY_OPTION_RIGHT].value.action);
+    removeScreenAction (s,
+	   		&rd->opt[ROTATE_DISPLAY_OPTION_LEFT_WINDOW].value.action);
+    removeScreenAction (s,
+	   		&rd->opt[ROTATE_DISPLAY_OPTION_RIGHT_WINDOW].value.action);
+    removeScreenAction (s, 
+			&rd->opt[ROTATE_DISPLAY_OPTION_FLIP_LEFT].value.action);
+    removeScreenAction (s,
+	   		&rd->opt[ROTATE_DISPLAY_OPTION_FLIP_RIGHT].value.action);
 
     UNWRAP (rs, s, preparePaintScreen);
     UNWRAP (rs, s, donePaintScreen);
