@@ -33,7 +33,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/poll.h>
-#include <unistd.h>
 
 #define XK_MISCELLANY
 #include <X11/keysymdef.h>
@@ -84,8 +83,6 @@ static int        targetOutput = 0;
 
 static Bool inHandleEvent = FALSE;
 
-static Bool shutDown = FALSE;
-
 static const CompTransform identity = {
     {
 	1.0, 0.0, 0.0, 0.0,
@@ -394,9 +391,9 @@ showDesktop (CompDisplay     *d,
     if (s)
     {
 	if (s->showingDesktopMask == 0)
-	    enterShowDesktopMode (s);
+	    (*s->enterShowDesktopMode) (s);
 	else
-	    leaveShowDesktopMode (s, NULL);
+	    (*s->leaveShowDesktopMode) (s, NULL);
     }
 
     return TRUE;
@@ -422,11 +419,11 @@ toggleSlowAnimations (CompDisplay     *d,
 }
 
 static Bool
-raise (CompDisplay     *d,
-       CompAction      *action,
-       CompActionState state,
-       CompOption      *option,
-       int	       nOption)
+raiseInitiate (CompDisplay     *d,
+	       CompAction      *action,
+	       CompActionState state,
+	       CompOption      *option,
+	       int	       nOption)
 {
     CompWindow *w;
     Window     xid;
@@ -441,11 +438,11 @@ raise (CompDisplay     *d,
 }
 
 static Bool
-lower (CompDisplay     *d,
-       CompAction      *action,
-       CompActionState state,
-       CompOption      *option,
-       int	       nOption)
+lowerInitiate (CompDisplay     *d,
+	       CompAction      *action,
+	       CompActionState state,
+	       CompOption      *option,
+	       int	       nOption)
 {
     CompWindow *w;
     Window     xid;
@@ -471,25 +468,22 @@ changeWindowOpacity (CompWindow *w,
     if (w->type & CompWindowTypeDesktopMask)
 	return;
 
-    step = (OPAQUE * w->screen->opacityStep) / 100;
+    step = (0xff * w->screen->opacityStep) / 100;
 
-    opacity = w->paint.opacity + step * direction;
-    if (opacity > OPAQUE)
+    w->opacityFactor = w->opacityFactor + step * direction;
+    if (w->opacityFactor > 0xff)
     {
-	opacity = OPAQUE;
+	w->opacityFactor = 0xff;
     }
-    else if (opacity < step)
+    else if (w->opacityFactor < step)
     {
-	opacity = step;
+	w->opacityFactor = step;
     }
 
-    if (w->paint.opacity != opacity)
+    opacity = (w->opacity * w->opacityFactor) / 0xff;
+    if (opacity != w->paint.opacity)
     {
 	w->paint.opacity = opacity;
-
-	setWindowProp32 (w->screen->display, w->id,
-			 w->screen->display->winOpacityAtom,
-			 w->paint.opacity);
 	addWindowDamage (w);
     }
 }
@@ -1014,7 +1008,7 @@ compDisplayInitOptions (CompDisplay *display,
     o->shortDesc		     = N_("Raise Window");
     o->longDesc			     = N_("Raise window above other windows");
     o->type			     = CompOptionTypeAction;
-    o->value.action.initiate	     = raise;
+    o->value.action.initiate	     = raiseInitiate;
     o->value.action.terminate        = 0;
     o->value.action.bell	     = FALSE;
     o->value.action.edgeMask	     = 0;
@@ -1029,7 +1023,7 @@ compDisplayInitOptions (CompDisplay *display,
     o->shortDesc		     = N_("Lower Window");
     o->longDesc			     = N_("Lower window beneath other windows");
     o->type			     = CompOptionTypeAction;
-    o->value.action.initiate	     = lower;
+    o->value.action.initiate	     = lowerInitiate;
     o->value.action.terminate        = 0;
     o->value.action.bell	     = FALSE;
     o->value.action.edgeMask	     = 0;
@@ -2053,14 +2047,10 @@ eventLoop (void)
 	if (display->dirtyPluginList)
 	    updatePlugins (display);
 
-	if (restartSignal)
+	if (restartSignal || shutDown)
 	{
-	    execvp (programName, programArgv);
-	    exit (1);
-	}
-	else if (shutDown)
-	{
-	    exit (0);
+	    while (popPlugin ());
+	    return;
 	}
 
 	while (XPending (display->display))
@@ -2536,7 +2526,9 @@ addDisplay (char *name,
     Window	focus;
     int		revertTo, i;
     int		compositeMajor, compositeMinor;
+    int		fixesMinor;
     int		xkbOpcode;
+    int		firstScreen, lastScreen;
 
     d = &compDisplay;
 
@@ -2608,6 +2600,10 @@ addDisplay (char *name,
 
     d->fileWatch = NULL;
 
+    d->matchInitExp	      = matchInitExp;
+    d->matchExpHandlerChanged = matchExpHandlerChanged;
+    d->matchPropertyChanged   = matchPropertyChanged;
+
     d->supportedAtom	     = XInternAtom (dpy, "_NET_SUPPORTED", 0);
     d->supportingWmCheckAtom = XInternAtom (dpy, "_NET_SUPPORTING_WM_CHECK", 0);
 
@@ -2813,6 +2809,21 @@ addDisplay (char *name,
 	return FALSE;
     }
 
+    if (!XFixesQueryExtension (dpy, &d->fixesEvent, &d->fixesError))
+    {
+	fprintf (stderr, "%s: No fixes extension\n", programName);
+	return FALSE;
+    }
+
+    XFixesQueryVersion (dpy, &d->fixesVersion, &fixesMinor);
+    /*
+    if (d->fixesVersion < 5)
+    {
+	fprintf (stderr, "%s: Need fixes extension version 5 or later "
+		 "for client-side cursor\n", programName);
+    }
+    */
+
     d->shapeExtension = XShapeQueryExtension (dpy,
 					      &d->shapeEvent,
 					      &d->shapeError);
@@ -2854,7 +2865,18 @@ addDisplay (char *name,
     d->escapeKeyCode = XKeysymToKeycode (dpy, XStringToKeysym ("Escape"));
     d->returnKeyCode = XKeysymToKeycode (dpy, XStringToKeysym ("Return"));
 
-    for (i = 0; i < ScreenCount (dpy); i++)
+    if (onlyCurrentScreen)
+    {
+	firstScreen = DefaultScreen (dpy);
+	lastScreen  = DefaultScreen (dpy);
+    }
+    else
+    {
+	firstScreen = 0;
+	lastScreen  = ScreenCount (dpy) - 1;
+    }
+
+    for (i = firstScreen; i <= lastScreen; i++)
     {
 	Window		     newWmSnOwner = None, newCmSnOwner = None;
 	Atom		     wmSnAtom = 0, cmSnAtom = 0;
@@ -3113,7 +3135,7 @@ focusDefaultWindow (CompDisplay *d)
 				   CompWindowTypeDialogMask |
 				   CompWindowTypeModalDialogMask))
 		    {
-			if (w->activeNum > focus->activeNum)
+			if (compareWindowActiveness (focus, w) < 0)
 			    focus = w;
 		    }
 		}
@@ -3599,3 +3621,15 @@ fileWatchRemoved (CompDisplay   *display,
 		  CompFileWatch *fileWatch)
 {
 }
+
+CompCursor *
+findCursorAtDisplay (CompDisplay *display)
+{
+    CompScreen *s;
+
+    for (s = display->screens; s; s = s->next)
+	if (s->cursors)
+	    return s->cursors;
+
+    return NULL;
+}
