@@ -30,6 +30,7 @@
 #include <X11/Xatom.h>
 #include <X11/extensions/shape.h>
 #include <X11/extensions/Xrandr.h>
+#include <X11/extensions/Xfixes.h>
 
 #include <compiz.h>
 
@@ -46,7 +47,7 @@ handleWindowDamageRect (CompWindow *w,
     REGION region;
     Bool   initial = FALSE;
 
-    if (!w->redirected)
+    if (!w->redirected || w->bindFailed)
 	return;
 
     if (!w->damaged)
@@ -70,14 +71,14 @@ handleWindowDamageRect (CompWindow *w,
 	region.rects = &region.extents;
 	region.numRects = region.size = 1;
 
-	damageWindowRegion (w, &region);
-
-	if (initial)
-	    damageWindowOutputExtents (w);
+	damageScreenRegion (w->screen, &region);
     }
 
     if (!w->attrib.override_redirect)
 	w->placed = TRUE;
+
+    if (initial)
+	damageWindowOutputExtents (w);
 }
 
 void
@@ -91,6 +92,8 @@ handleSyncAlarm (CompWindow *w)
 	    w->syncWaitHandle = 0;
 	}
 
+	w->syncWait = FALSE;
+
 	if (resizeWindow (w,
 			  w->syncX, w->syncY,
 			  w->syncWidth, w->syncHeight,
@@ -111,7 +114,12 @@ handleSyncAlarm (CompWindow *w)
 	    }
 
 	    w->nDamage = 0;
-	    w->syncWait = FALSE;
+	}
+	else
+	{
+	    /* resizeWindow failing means that there is another pending
+	       resize and we must send a new sync request to the client */
+	    sendSyncRequest (w);
 	}
     }
 }
@@ -153,7 +161,7 @@ moveInputFocusToOtherWindow (CompWindow *w)
 					   CompWindowTypeDialogMask |
 					   CompWindowTypeModalDialogMask))
 			    {
-				if (a->activeNum > focus->activeNum)
+				if (compareWindowActiveness (focus, a) < 0)
 				    focus = a;
 			    }
 			}
@@ -259,6 +267,32 @@ triggerButtonPressBindings (CompDisplay *d,
     CompAction	    *action;
     unsigned int    modMask = REAL_MOD_MASK & ~d->ignoredModMask;
     unsigned int    bindMods;
+    unsigned int    edge = 0;
+
+    if (edgeWindow)
+    {
+	CompScreen   *s;
+	unsigned int i;
+
+	s = findScreenAtDisplay (d, event->xbutton.root);
+	if (!s)
+	    return FALSE;
+
+	if (event->xbutton.window != edgeWindow)
+	{
+	    if (!s->maxGrab || event->xbutton.window != s->root)
+		return FALSE;
+	}
+
+	for (i = 0; i < SCREEN_EDGE_NUM; i++)
+	{
+	    if (edgeWindow == s->screenEdge[i].id)
+	    {
+		edge = 1 << i;
+		break;
+	    }
+	}
+    }
 
     while (nOption--)
     {
@@ -275,6 +309,22 @@ triggerButtonPressBindings (CompDisplay *d,
 	    }
 	}
 
+	if (edge)
+	{
+	    if (isInitiateBinding (option, CompBindingTypeEdgeButton, state,
+				   &action))
+	    {
+		if (action->edgeMask & edge)
+		{
+		    if (action->edgeButton == event->xbutton.button)
+			if ((*action->initiate) (d, action,
+						 CompActionStateInitEdge,
+						 argument, nArgument))
+			    return TRUE;
+		}
+	    }
+	}
+
 	option++;
     }
 
@@ -553,6 +603,9 @@ isEdgeEnterAction (CompOption      *option,
     if (!isEdgeAction (option, state, edge))
 	return FALSE;
 
+    if (option->value.action.type & CompBindingTypeEdgeButton)
+	return FALSE;
+
     if (!option->value.action.initiate)
 	return FALSE;
 
@@ -804,7 +857,7 @@ handleActionEvent (CompDisplay *d,
 
 		edgeWindow = None;
 
-		o[0].value.i = event->xcrossing.window;
+		o[0].value.i = d->activeWindow;
 		o[1].value.i = event->xcrossing.state;
 		o[2].value.i = event->xcrossing.x_root;
 		o[3].value.i = event->xcrossing.y_root;
@@ -848,7 +901,7 @@ handleActionEvent (CompDisplay *d,
 
 		edgeWindow = event->xcrossing.window;
 
-		o[0].value.i = event->xcrossing.window;
+		o[0].value.i = d->activeWindow;
 		o[1].value.i = event->xcrossing.state;
 		o[2].value.i = event->xcrossing.x_root;
 		o[3].value.i = event->xcrossing.y_root;
@@ -912,7 +965,7 @@ handleActionEvent (CompDisplay *d,
 	    {
 		state = CompActionStateTermEdgeDnd;
 
-		o[0].value.i = event->xclient.window;
+		o[0].value.i = d->activeWindow;
 		o[1].value.i = 0; /* fixme */
 		o[2].value.i = 0; /* fixme */
 		o[3].value.i = 0; /* fixme */
@@ -967,7 +1020,7 @@ handleActionEvent (CompDisplay *d,
 	    {
 		state = CompActionStateInitEdgeDnd;
 
-		o[0].value.i = xdndWindow;
+		o[0].value.i = d->activeWindow;
 		o[1].value.i = 0; /* fixme */
 		o[2].value.i = event->xclient.data.l[2] >> 16;
 		o[3].value.i = event->xclient.data.l[2] & 0xffff;
@@ -994,7 +1047,18 @@ handleActionEvent (CompDisplay *d,
 	}
 	break;
     default:
-	if (event->type == d->xkbEvent)
+	if (event->type == d->fixesEvent + XFixesCursorNotify)
+	{
+	    /*
+	    XFixesCursorNotifyEvent *ce = (XFixesCursorNotifyEvent *) event;
+	    CompCursor		    *cursor;
+
+	    cursor = findCursorAtDisplay (d);
+	    if (cursor)
+		updateCursor (cursor, ce->x, ce->y, ce->cursor_serial);
+	    */
+	}
+	else if (event->type == d->xkbEvent)
 	{
 	    XkbAnyEvent *xkbEvent = (XkbAnyEvent *) event;
 
@@ -1059,6 +1123,15 @@ handleActionEvent (CompDisplay *d,
     return FALSE;
 }
 
+void
+handleCompizEvent (CompDisplay *d,
+		   char        *pluginName,
+		   char        *eventName,
+		   CompOption  *option,
+		   int         nOption)
+{
+}
+
 void
 handleEvent (CompDisplay *d,
 	     XEvent      *event)
@@ -1074,6 +1147,13 @@ handleEvent (CompDisplay *d,
 						       event->xbutton.x_root,
 						       event->xbutton.y_root));
 	break;
+    case MotionNotify:
+	s = findScreenAtDisplay (d, event->xmotion.root);
+	if (s)
+	    setCurrentOutput (s, outputDeviceForPoint (s,
+						       event->xmotion.x_root,
+						       event->xmotion.y_root));
+	break;
     case KeyPress:
 	w = findWindowAtDisplay (d, d->activeWindow);
 	if (w)
@@ -1212,9 +1292,8 @@ handleEvent (CompDisplay *d,
 		    w->minimized = FALSE;
 		    w->state &= ~CompWindowStateHiddenMask;
 
-		    (*w->screen->windowStateChangeNotify) (w);
+		    changeWindowState (w, w->state);
 
-		    setWindowState (d, w->state, w->id);
 		    updateClientListForScreen (w->screen);
 		}
 
@@ -1281,25 +1360,7 @@ handleEvent (CompDisplay *d,
 	}
 	break;
     case PropertyNotify:
-	if (event->xproperty.atom == d->winActiveAtom)
-	{
-	    Window newActiveWindow;
-
-	    newActiveWindow = getActiveWindow (d, event->xproperty.window);
-	    if (newActiveWindow != d->activeWindow)
-	    {
-		d->activeWindow = newActiveWindow;
-
-		s = findScreenAtDisplay (d, event->xproperty.window);
-		if (s)
-		{
-		    w = findWindowAtDisplay (d, newActiveWindow);
-		    if (w)
-			w->activeNum = s->activeNum++;
-		}
-	    }
-	}
-	else if (event->xproperty.atom == d->winTypeAtom)
+	if (event->xproperty.atom == d->winTypeAtom)
 	{
 	    w = findWindowAtDisplay (d, event->xproperty.window);
 	    if (w)
@@ -1326,18 +1387,25 @@ handleEvent (CompDisplay *d,
 		    if (w->type & CompWindowTypeDesktopMask)
 			w->paint.opacity = OPAQUE;
 
+		    if (type & (CompWindowTypeDockMask |
+				CompWindowTypeDesktopMask))
+			setDesktopForWindow (w, 0xffffffff);
+
 		    updateClientListForScreen (w->screen);
+
+		    (*d->matchPropertyChanged) (d, w);
 		}
 	    }
 	}
 	else if (event->xproperty.atom == d->winStateAtom)
 	{
 	    w = findWindowAtDisplay (d, event->xproperty.window);
-	    if (w)
+	    if (w && !w->managed)
 	    {
 		unsigned int state;
 
 		state = getWindowState (d, w->id);
+		state = constrainWindowState (state, w->actions);
 
 		if (state != w->state)
 		{
@@ -1348,6 +1416,8 @@ handleEvent (CompDisplay *d,
 
 		    if (w->type & CompWindowTypeDesktopMask)
 			w->paint.opacity = OPAQUE;
+
+		    (*d->matchPropertyChanged) (d, w);
 		}
 	    }
 	}
@@ -1383,16 +1453,11 @@ handleEvent (CompDisplay *d,
 	    w = findWindowAtDisplay (d, event->xproperty.window);
 	    if (w && (w->type & CompWindowTypeDesktopMask) == 0)
 	    {
-		GLushort opacity;
+		w->opacity	  = OPAQUE;
+		w->opacityPropSet =
+		    readWindowProp32 (d, w->id, d->winOpacityAtom, &w->opacity);
 
-		opacity = getWindowProp32 (d, w->id,
-					   d->winOpacityAtom,
-					   OPAQUE);
-		if (opacity != w->opacity)
-		{
-		    w->opacity = w->paint.opacity = opacity;
-		    addWindowDamage (w);
-		}
+		updateWindowOpacity (w);
 	    }
 	}
 	else if (event->xproperty.atom == d->winBrightnessAtom)
@@ -1446,7 +1511,11 @@ handleEvent (CompDisplay *d,
 		finiTexture (s, &s->backgroundTexture);
 		initTexture (s, &s->backgroundTexture);
 
-		damageScreen (s);
+		if (s->backgroundLoaded)
+		{
+		    s->backgroundLoaded = FALSE;
+		    damageScreen (s);
+		}
 	    }
 	}
 	else if (event->xproperty.atom == d->wmStrutAtom ||
@@ -1492,13 +1561,6 @@ handleEvent (CompDisplay *d,
 		w->startupId = getStartupId (w);
 	    }
 	}
-	else if (event->xproperty.atom == d->winDesktopAtom)
-	{
-	    if (getWindowProp32 (d, event->xproperty.window,
-				 d->winDesktopAtom, 1) != 0)
-		setWindowProp32 (d, event->xproperty.window,
-				 d->winDesktopAtom, 0);
-	}
 	else if (event->xproperty.atom == XA_WM_CLASS)
 	{
 	    w = findWindowAtDisplay (d, event->xproperty.window);
@@ -1520,17 +1582,9 @@ handleEvent (CompDisplay *d,
 	    w = findWindowAtDisplay (d, event->xclient.window);
 	    if (w && (w->type & CompWindowTypeDesktopMask) == 0)
 	    {
-		GLushort opacity;
+		GLushort opacity = event->xclient.data.l[0] >> 16;
 
-		opacity = event->xclient.data.l[0] >> 16;
-		if (opacity != w->paint.opacity)
-		{
-		    w->paint.opacity = opacity;
-
-		    setWindowProp32 (d, w->id, d->winOpacityAtom,
-				     w->paint.opacity);
-		    addWindowDamage (w);
-		}
+		setWindowProp32 (d, w->id, d->winOpacityAtom, opacity);
 	    }
 	}
 	else if (event->xclient.message_type == d->winBrightnessAtom)
@@ -1538,17 +1592,9 @@ handleEvent (CompDisplay *d,
 	    w = findWindowAtDisplay (d, event->xclient.window);
 	    if (w)
 	    {
-		GLushort brightness;
+		GLushort brightness = event->xclient.data.l[0] >> 16;
 
-		brightness = event->xclient.data.l[0] >> 16;
-		if (brightness != w->paint.brightness)
-		{
-		    w->paint.brightness = brightness;
-
-		    setWindowProp32 (d, w->id, d->winBrightnessAtom,
-				     w->paint.brightness);
-		    addWindowDamage (w);
-		}
+		setWindowProp32 (d, w->id, d->winBrightnessAtom, brightness);
 	    }
 	}
 	else if (event->xclient.message_type == d->winSaturationAtom)
@@ -1556,22 +1602,9 @@ handleEvent (CompDisplay *d,
 	    w = findWindowAtDisplay (d, event->xclient.window);
 	    if (w && w->screen->canDoSaturated)
 	    {
-		GLushort saturation;
+		GLushort saturation = event->xclient.data.l[0] >> 16;
 
-		saturation = event->xclient.data.l[0] >> 16;
-		if (saturation != w->saturation)
-		{
-		    w->saturation = saturation;
-
-		    setWindowProp32 (d, w->id, d->winSaturationAtom,
-				     w->saturation);
-
-		    if (w->alive)
-		    {
-			w->paint.saturation = w->saturation;
-			addWindowDamage (w);
-		    }
-		}
+		setWindowProp32 (d, w->id, d->winSaturationAtom, saturation);
 	    }
 	}
 	else if (event->xclient.message_type == d->winStateAtom)
@@ -1617,11 +1650,9 @@ handleEvent (CompDisplay *d,
 		    recalcWindowType (w);
 		    recalcWindowActions (w);
 
-		    (*w->screen->windowStateChangeNotify) (w);
+		    changeWindowState (w, w->state);
 
 		    updateWindowAttributes (w, FALSE);
-
-		    setWindowState (d, w->state, w->id);
 		}
 	    }
 	}
@@ -1672,7 +1703,11 @@ handleEvent (CompDisplay *d,
 
 		value.i = event->xclient.data.l[0] / s->width;
 
-		(*s->setScreenOption) (s, "size", &value);
+		(*s->setScreenOption) (s, "hsize", &value);
+
+		value.i = event->xclient.data.l[1] / s->height;
+
+		(*s->setScreenOption) (s, "vsize", &value);
 	    }
 	}
 	else if (event->xclient.message_type == d->moveResizeWindowAtom)
@@ -1762,12 +1797,36 @@ handleEvent (CompDisplay *d,
 		    event->xclient.window == None)
 		{
 		    if (event->xclient.data.l[0])
-			enterShowDesktopMode (s);
+			(*s->enterShowDesktopMode) (s);
 		    else
-			leaveShowDesktopMode (s, NULL);
+			(*s->leaveShowDesktopMode) (s, NULL);
 		}
 	    }
 	}
+	else if (event->xclient.message_type == d->numberOfDesktopsAtom)
+	{
+	    s = findScreenAtDisplay (d, event->xclient.window);
+	    if (s)
+	    {
+		CompOptionValue value;
+
+		value.i = event->xclient.data.l[0];
+
+		(*s->setScreenOption) (s, "number_of_desktops", &value);
+	    }
+	}
+	else if (event->xclient.message_type == d->currentDesktopAtom)
+	{
+	    s = findScreenAtDisplay (d, event->xclient.window);
+	    if (s)
+		setCurrentDesktop (s, event->xclient.data.l[0]);
+	}
+	else if (event->xclient.message_type == d->winDesktopAtom)
+	{
+	    w = findWindowAtDisplay (d, event->xclient.window);
+	    if (w)
+		setDesktopForWindow (w, event->xclient.data.l[0]);
+	}
 	break;
     case MappingNotify:
 	updateModifierMappings (d);
@@ -1779,7 +1838,7 @@ handleEvent (CompDisplay *d,
 	    if (w->minimized)
 		unminimizeWindow (w);
 
-	    leaveShowDesktopMode (w->screen, w);
+	    (*w->screen->leaveShowDesktopMode) (w->screen, w);
 
 	    w->managed = TRUE;
 
@@ -1788,15 +1847,35 @@ handleEvent (CompDisplay *d,
 		w->initialViewportX = w->screen->x;
 		w->initialViewportY = w->screen->y;
 
+		w->initialTimestampSet = FALSE;
+
 		applyStartupProperties (w->screen, w);
 
+		w->pendingMaps++;
+
 		XMapWindow (d->display, event->xmaprequest.window);
 
 		updateWindowAttributes (w, FALSE);
 
 		if (focusWindowOnMap (w))
+		{
 		    moveInputFocusToWindow (w);
+		}
+		else if (w->type & ~CompWindowTypeSplashMask)
+		{
+		    CompWindow *p;
+
+		    for (p = w->prev; p; p = p->prev)
+			if (p->id == d->activeWindow)
+			    break;
+
+		    /* window is above active window so we should lower it */
+		    if (p)
+			restackWindowBelow (w, p);
+		}
 	    }
+
+	    setWindowProp (d, w->id, d->winDesktopAtom, w->desktop);
 	}
 	else
 	{
@@ -1845,27 +1924,35 @@ handleEvent (CompDisplay *d,
     case FocusIn:
 	if (event->xfocus.mode != NotifyGrab)
 	{
-	    w = findWindowAtDisplay (d, event->xfocus.window);
+	    w = findTopLevelWindowAtDisplay (d, event->xfocus.window);
 	    if (w && w->managed)
 	    {
 		unsigned int state = w->state;
 
 		if (w->id != d->activeWindow)
+		{
+		    d->activeWindow = w->id;
+		    w->activeNum = w->screen->activeNum++;
+
+		    addToCurrentActiveWindowHistory (w->screen, w->id);
+
 		    XChangeProperty (d->display, w->screen->root,
 				     d->winActiveAtom,
 				     XA_WINDOW, 32, PropModeReplace,
 				     (unsigned char *) &w->id, 1);
+		}
 
-		w->state &= ~CompWindowStateDemandsAttentionMask;
+		state &= ~CompWindowStateDemandsAttentionMask;
 
 		if (w->state != state)
-		    setWindowState (d, w->state, w->id);
+		    changeWindowState (w, state);
 	    }
 	}
 	break;
     case EnterNotify:
 	if (!d->screens->maxGrab		    &&
 	    event->xcrossing.mode   != NotifyGrab   &&
+	    event->xcrossing.mode   != NotifyUngrab &&
 	    event->xcrossing.detail != NotifyInferior)
 	{
 	    Bool raise;
